"""Test module for testing embeddings function"""
import numpy as np
import pytest

test_params = [
    (1, ["cls", "mean"]),
    (2, ["full", "mean", "cls"]),
    (10, ["cls", "full"]),
]


@pytest.mark.parametrize("batch_size, pool_mode", test_params)
def test_embeddings_type_and_shape(init_model, sequences, batch_size, pool_mode):
    test_trans = init_model
    embeddings = test_trans.compute_embeddings(
        sequences,
        batch_size=batch_size,
        pool_mode=pool_mode,
    )

    assert isinstance(embeddings, dict)
    if "full" in pool_mode:
        for emb, sequence in zip(embeddings["full"], sequences):
            assert emb.shape[0] == len(sequence)
    if "cls" in pool_mode:
        assert isinstance(embeddings["cls"], np.ndarray)

    if "mean" in pool_mode:
        assert isinstance(embeddings["mean"], np.ndarray)


@pytest.mark.parametrize("batch_size, pool_mode", test_params)
def test_embeddings_type_and_shape_fasta(
    init_model, fasta_path, lengths_sequence_fasta, batch_size, pool_mode
):
    test_trans = init_model
    embeddings = test_trans.compute_embeddings(
        fasta_path,
        batch_size=batch_size,
        pool_mode=pool_mode,
    )
    if "full" in pool_mode:
        for emb, length in zip(embeddings["full"], lengths_sequence_fasta):
            assert emb.shape[0] == length

    assert isinstance(embeddings, dict)
    if "cls" in pool_mode:
        assert isinstance(embeddings["cls"], np.ndarray)

    if "mean" in pool_mode:
        assert isinstance(embeddings["mean"], np.ndarray)



"""Test module for testing loglikelihoods function"""
import pytest
from numpy.testing import assert_allclose

test_params = [
    ([["A1Q"], ["A1K", "K2A"], ["A1H"], ["K3W", "K2D", "L9H"]], "params1"),
]


@pytest.mark.parametrize("mutations, params", test_params)
def test_mutation_score_type_shape_and_range(
    init_model, sequences, mutations_score_results, mutations, params
):
    test_trans = init_model
    mutations_scores = test_trans.compute_mutation_score(sequences, mutations)
    assert len(mutations_scores) == len(sequences)
    if test_trans._model_dir in mutations_score_results.keys():
        results = mutations_score_results[test_trans._model_dir][params]
        assert_allclose(mutations_scores, results, rtol=0.01)



ESM_LIST = [
    # "esm1_t34_670M_UR50S",
    # "esm1_t34_670M_UR50D",
    "esm1_t34_670M_UR100",
    "esm1_t12_85M_UR50S",
    "esm1_t6_43M_UR50S",
    "esm1b_t33_650M_UR50S",
    "esm_msa1_t12_100M_UR50S",
    "esm_msa1b_t12_100M_UR50S",
    "esm1v_t33_650M_UR90S_1",
]

ROSTLAB_LIST = ["Rostlab/prot_bert", "Rostlab/prot_bert_bfd"]

MAPPING_PROTBERT = {
    "protbert": "Rostlab/prot_bert",
    "protbert_bfd": "Rostlab/prot_bert_bfd",
}

DEFAULT_ESM_MODEL = "esm1_t34_670M_UR100"
DEFAULT_ROSTLAB_MODEL = "Rostlab/prot_bert"

BACKEND_LIST = ESM_LIST + list(MAPPING_PROTBERT.keys())

NATURAL_AAS_LIST = list("ACDEFGHIKLMNPQRSTVWY")



from typing import Dict, List, Tuple

import pytorch_lightning as pl
import torch
import torchmetrics
from torch.nn import functional as F  # noqa: N812 pylint: disable=wrong-import-order

from .optimizer import lr_update


class LightningModule(pl.LightningModule):
    """Create lightning model to use ddp"""

    def __init__(
        self,
        model,
        alphabet,
        lr: float,
        warmup_end_lr: float,
        warmup_updates: int = 10,
        warmup_init_lr: float = 1e-7,
    ):
        super().__init__()
        self.model = model
        self.alphabet = alphabet
        self.lr = lr
        self.automatic_optimization = True
        self.warmup_updates = warmup_updates
        self.warmup_init_lr = min(warmup_init_lr, lr)
        self.lr_step = (warmup_end_lr - self.warmup_init_lr) / warmup_updates
        self.decay_factor = warmup_end_lr * warmup_updates ** 0.5
        self.train_acc = torchmetrics.Accuracy()
        self.val_acc = torchmetrics.Accuracy()

    def forward(self, x):
        return self.model(x)["logits"]

    def configure_optimizers(self) -> Tuple[List[torch.optim.Optimizer], List[Dict]]:
        """Configure the optimizer and learning rate scheduler.

        Returns:
            - list of optimizers.
            - list of lr schedulers.
        """
        optimizer = torch.optim.Adam(self.model.parameters(), self.lr)

        lr_scheduler = {
            "scheduler": torch.optim.lr_scheduler.LambdaLR(
                optimizer,
                lr_lambda=lambda x: lr_update(
                    num_updates=x,
                    warmup_updates=self.warmup_updates,
                    warmup_init_lr=self.warmup_init_lr,
                    lr_step=self.lr_step,
                    decay_factor=self.decay_factor,
                ),
            ),
            "name": "learning_rate",
            "interval": "step",
            "frequency": 1,
        }

        return [optimizer], [lr_scheduler]

    def cross_entropy_loss(self, logits, targets):
        return F.cross_entropy(
            logits.reshape(-1, logits.size(-1)),
            targets.reshape(-1),
            reduction="sum",
            ignore_index=self.alphabet.padding_idx,
        )

    def training_step(self, train_batch, batch_idx):
        tokens, target = train_batch
        logits = self.forward(tokens)
        loss = self.cross_entropy_loss(logits, target)

        masked_preds, masked_targets = self.get_tensor_accuracy(logits, target)
        self.train_acc(masked_preds, masked_targets)

        masked_tokens = target.ne(self.alphabet.padding_idx)
        sample_size = masked_tokens.int().sum()
        loss = loss / sample_size

        self.log_dict(
            {"train_loss": loss, "train_acc": self.train_acc},
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        return loss

    def validation_step(self, val_batch, batch_idx):
        """Log the loss and metrics for a batch.

        Args:
            batch: batch input.
            batch_idx: index of the batch.
        """
        tokens, target = val_batch
        logits = self.forward(tokens)
        loss = self.cross_entropy_loss(logits, target)

        masked_preds, masked_targets = self.get_tensor_accuracy(logits, target)
        self.val_acc(masked_preds, masked_targets)

        masked_tokens = target.ne(self.alphabet.padding_idx)
        sample_size = masked_tokens.int().sum()
        loss = loss / sample_size

        self.log_dict(
            {"val_loss": loss, "val_acc": self.val_acc},
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        return loss

    def get_tensor_accuracy(
        self, logits: torch.Tensor, targets: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Calculate accuracy for multi-masking, summed over batch.

        Args:
            logits: prediction from the model, shape = (batch, len_tokens, len_vocab)
            targets: ground truth, shape = (batch, len_tokens)

        Returns:
            accuracy value.
        """
        preds = torch.argmax(logits, dim=-1)  # (batch, len_tokens)
        masked_tokens = targets.ne(self.alphabet.padding_idx)

        masked_preds = torch.masked_select(preds, masked_tokens)
        masked_targets = torch.masked_select(targets, masked_tokens)

        return masked_preds.detach().cpu(), masked_targets.detach().cpu()



"""
This script defines a class which inherits from the LanguageModel class, and is
specific to the ESM model developed by FAIR (https://github.com/facebookresearch/esm).
"""

from typing import Dict, List, Tuple

import esm
import torch
from biotransformers.lightning_utils.data import (
    AlphabetDataLoader,
    convert_ckpt_to_statedict,
)
from biotransformers.utils.constant import DEFAULT_ESM_MODEL, ESM_LIST
from biotransformers.utils.logger import logger  # noqa
from biotransformers.utils.utils import _generate_chunks, _get_num_batch_iter
from biotransformers.wrappers.language_model import LanguageModel
from ray.actor import ActorHandle
from tqdm import tqdm

log = logger("esm_wrapper")
path_msa_folder = str


class ESMWrapper(LanguageModel):
    """
    Class that uses an ESM type of pretrained transformers model to evaluate
    a protein likelihood so as other insights.
    """

    def __init__(self, model_dir: str, device: str):
        if model_dir not in ESM_LIST:
            print(
                f"Model dir '{model_dir}' not recognized. Using '{DEFAULT_ESM_MODEL}' as default"
            )
            model_dir = DEFAULT_ESM_MODEL
        super().__init__(model_dir=model_dir, device=device)
        self._model, self.alphabet = esm.pretrained.load_model_and_alphabet(model_dir)
        self.num_layers = self._model.num_layers
        repr_layers = -1
        self.repr_layers = (repr_layers + self.num_layers + 1) % (self.num_layers + 1)
        self.hidden_size = self._model.args.embed_dim
        self._model = self._model.to(self._device)
        self.batch_converter = self.alphabet.get_batch_converter()
        self.is_msa = "msa" in model_dir

    @property
    def model(self) -> torch.nn.Module:
        """Return torch model."""
        return self._model

    @property
    def clean_model_id(self) -> str:
        """Clean model ID (in case the model directory is not)"""
        return self.model_id

    @property
    def model_vocabulary(self) -> List[str]:
        """Returns the whole vocabulary list"""
        return list(self.alphabet.tok_to_idx.keys())

    @property
    def vocab_size(self) -> int:
        """Returns the whole vocabulary size"""
        return len(list(self.alphabet.tok_to_idx.keys()))

    @property
    def mask_token(self) -> str:
        """Representation of the mask token (as a string)"""
        return self.alphabet.all_toks[self.alphabet.mask_idx]  # "<mask>"

    @property
    def pad_token(self) -> str:
        """Representation of the pad token (as a string)"""
        return self.alphabet.all_toks[self.alphabet.padding_idx]  # "<pad>"

    @property
    def begin_token(self) -> str:
        """Representation of the beginning of sentence token (as a string)"""
        return self.alphabet.all_toks[self.alphabet.cls_idx]  # "<cls>"

    @property
    def end_token(self) -> str:
        """Representation of the end of sentence token (as a string)"""
        return self.alphabet.all_toks[self.alphabet.eos_idx]  # "<eos>"

    @property
    def does_end_token_exist(self) -> bool:
        """Returns true if a end of sequence token exists"""
        return self.alphabet.append_eos

    @property
    def token_to_id(self):
        """Returns a function which maps tokens to IDs"""
        return lambda x: self.alphabet.tok_to_idx[x]

    @property
    def embeddings_size(self):
        """Returns size of the embeddings"""
        return self.hidden_size

    def process_sequences_and_tokens(
        self, sequences_list: List[str]
    ) -> Dict[str, torch.Tensor]:
        """Function to transform tokens string to IDs; it depends on the model used"""
        if self.is_msa:
            _, _, all_tokens = self.batch_converter(sequences_list)
        else:
            _, _, all_tokens = self.batch_converter(
                [("", sequence) for sequence in sequences_list]
            )

        all_tokens = all_tokens.to("cpu")
        encoded_inputs = {
            "input_ids": all_tokens,
            "attention_mask": 1 * (all_tokens != self.token_to_id(self.pad_token)),
            "token_type_ids": torch.zeros(all_tokens.shape),
        }
        return encoded_inputs

    def _load_model(self, path_model: str, map_location=None):
        """Load model."""
        if path_model.endswith(".pt"):
            loaded_model = torch.load(path_model)
        elif path_model.endswith(".ckpt"):
            loaded_model = convert_ckpt_to_statedict(
                torch.load(path_model)["state_dict"]
            )
        else:
            raise ValueError("Expecting a .pt or .ckpt file")
        self._model.load_state_dict(loaded_model, map_location)
        self._model.eval()
        log.info("Load model %s" % path_model)

    def model_pass(
        self,
        model_inputs: Dict[str, torch.Tensor],
        batch_size: int,
        silent: bool = False,
        pba: ActorHandle = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Function which computes logits and embeddings based on a list of sequences,
        a provided batch size and an inference configuration. The output is obtained
        by computing a forward pass through the model ("forward inference")

        The datagenerator is not the same the multi_gpus inference. We use a tqdm progress bar
        that is updated by the worker. The progress bar is instantiated before ray.remote

        Args:
            model_inputs (Dict[str, torch.tensor]): [description]
            batch_size (int): size of the batch
            silent : display or not progress bar
            pba : tqdm progress bar for ray actor
        Returns:
            Tuple[torch.tensor, torch.tensor]:
                    * logits [num_seqs, max_len_seqs, vocab_size]
                    * embeddings [num_seqs, max_len_seqs+1, embedding_size]
        """
        if pba is None:
            batch_generator = tqdm(
                _generate_chunks(model_inputs, batch_size),
                total=_get_num_batch_iter(model_inputs, batch_size),
                disable=silent,
            )
        else:
            batch_generator = _generate_chunks(model_inputs, batch_size)

        logits = torch.Tensor()  # [num_seqs, max_len_seqs+1, vocab_size]
        embeddings = torch.Tensor()  # [num_seqs, max_len_seqs+1, embedding_size]
        for batch_inputs in batch_generator:
            with torch.no_grad():
                model_outputs = self._model(
                    batch_inputs["input_ids"].to(self._device),
                    repr_layers=[self.repr_layers],
                )
                batch_logits = model_outputs["logits"].detach().cpu()
                batch_embeddings = (
                    model_outputs["representations"][self.repr_layers].detach().cpu()
                )
                embeddings = torch.cat((embeddings, batch_embeddings), dim=0)
                logits = torch.cat((logits, batch_logits), dim=0)

            # tqdm worker update
            if pba is not None:
                pba.update.remote(1)
        return logits, embeddings

    def get_alphabet_dataloader(self):
        """Define an alphabet mapping for common method between
        protbert and ESM
        """

        def tokenize(x: List[str]):
            x_ = list(enumerate(x))
            _, seqs, tokens = self.batch_converter(x_)
            return seqs, tokens

        alphabet_dl = AlphabetDataLoader(
            prepend_bos=True,
            append_eos=True,
            mask_idx=self.alphabet.mask_idx,
            pad_idx=self.alphabet.padding_idx,
            standard_toks=self.alphabet.standard_toks,
            model_dir=self._model_dir,
            lambda_toks_to_ids=lambda x: self.alphabet.tok_to_idx[x],
            lambda_tokenizer=lambda x: tokenize(x),
        )

        return alphabet_dl



import functools
import random
from collections import OrderedDict
from typing import Callable, List, Sequence, Tuple

import numpy as np
import torch
from esm.data import BatchConverter
from torch.utils.data import DataLoader, Dataset, Sampler


class AlphabetDataLoader:
    """Class that carries tokenizer information"""

    def __init__(
        self,
        prepend_bos: bool,
        append_eos: bool,
        mask_idx: int,
        pad_idx: int,
        standard_toks: List[str],
        model_dir: str,
        lambda_toks_to_ids: Callable,
        lambda_tokenizer: Callable,
    ) -> None:
        self.prepend_bos = prepend_bos
        self.append_eos = append_eos
        self.mask_idx = mask_idx
        self.padding_idx = pad_idx
        self.standard_toks = standard_toks
        self.model_dir = model_dir
        self.lambda_toks_to_ids = lambda_toks_to_ids
        self.lambda_tokenizer = lambda_tokenizer

    def tok_to_idx(self, x):
        return self.lambda_toks_to_ids(x)

    def tokenizer(self):
        """Return seq-token based on sequence"""
        return self.lambda_tokenizer


def convert_ckpt_to_statedict(checkpoint_state_dict: OrderedDict) -> OrderedDict:
    """This function convert a state_dict coming form pytorch lightning checkpoint to
    a state_dict model that can be load directly in the bio-transformers model.

    The keys are updated so that it  m.jionatches those in the bio-transformers

    Args:
        checkpoint_state_dict: a state_dict loaded from a checkpoint
    """
    new_state_dict = OrderedDict()
    for k, v in checkpoint_state_dict.items():
        new_k = ".".join(k.split(".")[1:])  # remove model. prefix in key
        new_state_dict[new_k] = v.to("cpu")  # move tensor to cpu

    return new_state_dict


def worker_init_fn(worker_id: int):
    """Set numpy random seed for each worker.

    https://github.com/pytorch/pytorch/issues/5059#issuecomment-404232359

    Args:
        worker_id: unique id for each worker
    """
    np.random.seed(np.random.get_state()[1][0] + worker_id)


def mask_seq(
    seq: str,
    tokens: torch.Tensor,
    prepend_bos: bool,
    mask_idx: int,
    pad_idx: int,
    masking_ratio: float,
    masking_prob: float,
    random_token_prob: float,
    random_token_indices: List[int],
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Mask one sequence randomly.

    Args:
        seq: string of the sequence.
        tokens: tokens corresponding to the sequence, length can be longer than the seq.
        prepend_bos: if tokenizer adds <bos> token
        mask_idx: index of the mask token
        pad_idx: index of the padding token
        masking_ratio: ratio of tokens to be masked.
        masking_prob: probability that the chose token is replaced with a mask token.
        random_token_prob: probability that the chose token is replaced with a random token.
        random_token_indices: list of token indices that random replacement selects from.

    Returns:
        tokens: masked tokens
        targets: same length as tokens
    """
    # init
    seq_len = len(seq)
    mask_num = int(np.ceil(seq_len * masking_ratio))
    targets = tokens.detach().clone()
    # sample indices
    mask_indices = sorted(
        np.random.choice(seq_len, mask_num, replace=False) + int(prepend_bos)
    )
    # mask tokens
    for idx in mask_indices:
        rand = np.random.random()

        # replace with mask
        if rand < masking_prob:
            tokens[idx] = mask_idx

        # replace with random token
        elif rand < masking_prob + random_token_prob:
            tokens[idx] = np.random.choice(random_token_indices, 1)[0]

    # generate targets
    non_mask_indices = [i for i in range(seq_len) if i not in mask_indices]
    targets[non_mask_indices] = pad_idx

    return tokens, targets


def collate_fn(
    samples: Sequence[Tuple[str, str]],
    tokenizer: BatchConverter,
    alphabet: AlphabetDataLoader,
    masking_ratio: float,
    masking_prob: float,
    random_token_prob: float,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Collate function to mask tokens.

    Args:
        samples: a sequences of (label, seq).
        tokenizer: facebook tokenizer, that accepts sequences of (label, seq_str)
            and outputs (labels, seq_strs, tokens).
        alphabet: facebook alphabet.
        masking_ratio: ratio of tokens to be masked.
        masking_prob: probability that the chose token is replaced with a mask token.
        random_token_prob: probability that the chose token is replaced with a random token.

    Returns:
        tokens: model input
        targets: model target
        mask_indices: indices of masked tokens
    """
    random_token_indices = [alphabet.tok_to_idx(aa) for aa in alphabet.standard_toks]
    seqs, tokens = tokenizer(
        samples[0]
    )  # take samples[0] because batch_sampler return list of list
    tokens_list, targets_list = [], []
    for i, seq in enumerate(seqs):
        tokens_i, targets_i = mask_seq(
            seq=seq,
            tokens=tokens[i, :],
            prepend_bos=alphabet.prepend_bos,
            mask_idx=alphabet.mask_idx,
            pad_idx=alphabet.padding_idx,
            masking_ratio=masking_ratio,
            masking_prob=masking_prob,
            random_token_prob=random_token_prob,
            random_token_indices=random_token_indices,
        )
        tokens_list.append(tokens_i)
        targets_list.append(targets_i)

    tokens = torch.stack(tokens_list)
    targets = torch.stack(targets_list)

    return tokens, targets


def crop_sequence(sequence: str, crop_length: int) -> str:
    """If the length of the sequence is superior to crop_length, crop randomly
    the sequence to get the proper length."""
    if len(sequence) <= crop_length:
        return sequence
    else:
        start_idx = random.randint(0, len(sequence) - crop_length)
        return sequence[start_idx : (start_idx + crop_length)]


def get_batch_indices(
    sequence_strs: List[str],
    toks_per_batch: int,
    crop_sizes: Tuple[int, int] = (600, 1200),
) -> List[List[List[Tuple[int, int]]]]:
    """
    This sampler aims to create batches that do not contain fixed number of sequences
    but rather constant number of tokens. Some the batch can contain a few long
    sequences or multiple small ones.


    This sampler returns batches of indices to achieve this property. It also decides
    if sequences must be cropped and return the desired length. The cropping length is
    sampled randomly for each sequence at each epoch in the range of crop_sizes values.

    THis sampler computes a list of list of tuple which contains indices and
    lengths of sequences  inside the batch.
    Example:
        returning [[(1, 100), (3, 600)],[(4, 100), (7, 1200), (10, 600)], [(12, 1000)]]
        means that the first batch  will be composed of sequence at index 1 and 8 with
        lengths 100 and  600. The third batch contains only sequence 12 with a length
        of 1000.

    Args:
        sequence_strs: list of string
        toks_per_batch (int): Maximum number of token per batch
        extra_toks_per_seq (int, optional): . Defaults to 0.
        crop_sizes (Tuple[int, int]): min and max sequence lengths when cropping

    Returns:
        List: List of batches indexes and lengths
    """
    min_size, max_size = crop_sizes
    buffer_type = List[Tuple[int, int]]
    sizes = [(len(s), i) for i, s in enumerate(sequence_strs)]
    random.shuffle(sizes)
    batches: List[List[buffer_type]] = []
    buffer: buffer_type = []

    def _flush_current_buf():
        nonlocal buffer
        if len(buffer) == 0:
            return
        batches.append([buffer])
        buffer = []

    for sz, i in sizes:
        crop_size = random.randint(min_size, max_size) - 2
        if sz > crop_size:
            seq_length = crop_size
        else:
            seq_length = sz
        if seq_length + sum([b[1] for b in buffer]) > toks_per_batch:
            _flush_current_buf()
        buffer.append((i, seq_length))

    _flush_current_buf()
    return batches


class BatchWithConstantNumberTokensSampler(Sampler):
    """
    Sampler that returns batches of sequences indices in the dataset so that to ensure
    not a fixed number of sequences per batch but rather a fixed number of tokens per
    batch. This sampler also takes into account that we may want to crop dynamically
    sequences when sampling and thus returns in addition to indices, desired cropping
    lengths to inform the dataloader.
    """

    def __init__(
        self,
        sequence_strs: List[str],
        toks_per_batch: int,
        crop_sizes: Tuple[int, int] = (512, 1024),
    ):
        Sampler.__init__(self, data_source=None)
        self._sequence_strs = sequence_strs
        self._toks_per_batch = toks_per_batch
        self._crop_sizes = crop_sizes
        self._init_batches = get_batch_indices(
            sequence_strs=sequence_strs,
            toks_per_batch=toks_per_batch,
            crop_sizes=crop_sizes,
        )

    def __len__(self):
        return len(self._init_batches)

    def __iter__(self):
        yield from get_batch_indices(
            sequence_strs=self._sequence_strs,
            toks_per_batch=self._toks_per_batch,
            crop_sizes=self._crop_sizes,
        )


class BatchWithConstantNumberTokensDataset(Dataset):
    """
    Dataset class to work in pair with the BatchWithConstantNumberTokensSampler.
    """

    def __init__(self, sequences: List[str]):
        super().__init__()
        self.sequences = sequences

    def __len__(self):
        return len(self.sequences)

    def __getitem__(self, sampler_out) -> List[str]:
        indices = [out[0] for out in sampler_out]
        lengths = [out[1] for out in sampler_out]
        sequences = [
            crop_sequence(self.sequences[i], length)
            for i, length in zip(indices, lengths)
        ]
        return sequences


def create_dataloader(
    sequences: List[str],
    alphabet: AlphabetDataLoader,
    masking_ratio: float,
    masking_prob: float,
    random_token_prob: float,
    num_workers: int,
    toks_per_batch: int,
    crop_sizes: Tuple[int, int] = (512, 1024),
) -> DataLoader:
    """Create the PyTorch Dataloader.

    Args:
        filenames: list of sequences
        alphabet: facebook alphabet.
        filter_len: whether filter data wrt len.batch_seq
        num_workers: num of parallel data samplers
        masking_ratio: ratio of tokens to be masked.
        masking_prob: probability that the chose token is replaced with a mask token.
        random_token_prob: probability that the chose token is replaced with a random token.
        toks_per_batch: number of tokens per batch
        crop_sizes: range of values to crop dynamically sequences when sampling them

    Returns:
        torch DataLoader
    """

    dataset = BatchWithConstantNumberTokensDataset(sequences)
    batch_sampler = BatchWithConstantNumberTokensSampler(
        sequence_strs=sequences, toks_per_batch=toks_per_batch, crop_sizes=crop_sizes
    )

    loader = DataLoader(
        dataset,
        num_workers=num_workers,
        collate_fn=functools.partial(
            collate_fn,
            tokenizer=alphabet.tokenizer(),
            alphabet=alphabet,
            masking_ratio=masking_ratio,
            masking_prob=masking_prob,
            random_token_prob=random_token_prob,
        ),
        pin_memory=True,
        worker_init_fn=worker_init_fn,
        batch_sampler=batch_sampler,
        sampler=None,
    )
    return loader



import itertools
import os
import string
from glob import glob
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from Bio import SeqIO


def get_translation() -> Dict[int, Any]:
    """
    get translation dict to convert unused character in MSA
    """
    delete_keys = dict.fromkeys(string.ascii_lowercase)
    delete_keys["."] = None  # gap character  '-' alignement character
    delete_keys["*"] = None
    translation = str.maketrans(delete_keys)
    return translation


def read_sequence(filename: str) -> Tuple[str, str]:
    """Reads the first (reference) sequences from a fasta or MSA file."""
    record = next(SeqIO.parse(filename, "fasta"))
    return record.description, str(record.seq)


def remove_insertions(sequence: str) -> str:
    """Removes any insertions into the sequence.
    Needed to load aligned sequences in an MSA."""
    translation = get_translation()
    return sequence.translate(translation)


def read_msa(filename: str, nseq: int) -> List[Tuple[str, str]]:
    """Reads the first nseq sequences from an MSA file,
    automatically removes insertions."""
    return [
        (record.description, remove_insertions(str(record.seq)))
        for record in itertools.islice(SeqIO.parse(filename, "fasta"), nseq)
    ]


def get_msa_list(path_msa: Optional[str]) -> List[str]:
    """Get all files of the msa folder and check file format

    Args:
        path_msa (Optional[str]): path of the folder with a3m file
    """
    if path_msa is None:
        raise ValueError("The path of the msa folder could not be None with msa model.")
    if not os.path.isdir(path_msa):
        raise FileExistsError(f"{path_msa} is not a valid directory")

    list_msa = glob(path_msa + "/*.a3m")
    all_a3m_file = all([msa.endswith("a3m") for msa in list_msa])
    if len(list_msa) == 0:
        raise FileNotFoundError(
            "Can't find any msa files with .a3m format in this folder."
        )
    if not all_a3m_file:
        raise ValueError("All files in msa folder should have a3m format.")

    return list_msa


def get_msa_lengths(list_msa: List[List[Tuple[str, str]]], nseq: int) -> List[int]:
    """Get length of an MSA list

    All MSA must have at least nseq in msa

    Args:
        list_msa (List[List[Tuple[str,str]]]): list of MSA. MSA is a list of tuple
        nseq
    Returns:
        List[int]: [description]
    """

    def _msa_length(msa: List[Tuple[str, str]]) -> List[int]:
        """get length of each sequence in msa

        Example: >> input = ['AAAB','AAAA','AAA-']
                 >> _msa_length(input)
                 >> [4,4,4]
        Raises:
            ValueError if number of seq in the MSA is less than nseq
        Args:
            msa (List[Tuple[str, str]]): List of sequence

        Returns:
            List[int]: List of length of each msa
        """
        return [len(seq[1]) for seq in msa]

    lengths = [_msa_length(msa) for msa in list_msa]
    n_different_seq = sum([len(length) != nseq for length in lengths])
    if n_different_seq > 0:
        msg = (
            f"Find {n_different_seq} files with less than {nseq} sequences in the msa. "
            f"All msa files must have at least {nseq} sequences. "
            f"Use `from biotransformers.utils.msa_utils.msa_to_remove` to get the file to remove."
        )
        raise ValueError(msg)
    unique_length = [max(length) for length in lengths]
    return unique_length


def msa_to_remove(path_msa: str, n_seq) -> List[str]:
    """Get list of msa with less than nseq sequence

    Args:
        path_msa (str): [description]

    Returns:
        List of msa filepath that don't have enough enough sequences.
    """
    path_msa = str(Path(path_msa).resolve())
    list_msa_filepath = get_msa_list(path_msa)
    list_msa = [read_msa(file, n_seq) for file in list_msa_filepath]

    def _msa_length(msa: List[Tuple[str, str]]) -> List[int]:
        return [len(seq[1]) for seq in msa]

    lengths = [_msa_length(msa) for msa in list_msa]
    msa_to_remove = []
    for i, length in enumerate(lengths):
        if len(length) != n_seq:
            msa_to_remove.append(list_msa_filepath[i])
    print(
        f"{len(msa_to_remove)}/{len(list_msa)} have insufficient number of sequences in MSA."
    )
    return msa_to_remove



from biotransformers.bio_transformers import BioTransformers  # noqa
from biotransformers.utils.logger import logger  # noqa

from .version import VERSION  # noqa

log = logger("biotransformers")

__version__ = VERSION



"""Main module to build either ESM or protbert model"""

from biotransformers.utils.constant import BACKEND_LIST, MAPPING_PROTBERT
from biotransformers.utils.deprecated import deprecated_alias
from biotransformers.utils.utils import format_backend
from biotransformers.wrappers.esm_wrappers import ESMWrapper
from biotransformers.wrappers.rostlab_wrapper import RostlabWrapper
from biotransformers.wrappers.transformers_wrappers import TransformersWrapper


class BioTransformers(TransformersWrapper):
    """
    General class to choose an ESM or ProtBert backend
    Abstract method are implemented in transformers
    """

    def __init__(
        self,
        backend: str = "esm1_t6_43M_UR50S",
        num_gpus: int = 0,
    ):
        """General class to compute method for a list of provided backend

        If you want to restrict the use of GPUS, do make gpu1 and gpu3 available:
        os.environ["CUDA_VISIBLE_DEVICES"]="0,3" or export CUDA_VISIBLE_DEVICES="0,3"

        Args:
            backend (str, optional): name of the backend displayed with `list_backend()` . Defaults to "esm1_t6_43M_UR50S".
            num_gpus (int, optional): number of gpu to use. Defaults to 0.
        """
        pass

    @deprecated_alias(device="num_gpus")
    @deprecated_alias(multi_gpu="num_gpus")
    def __new__(
        cls,
        backend: str = "esm1_t6_43M_UR50S",
        num_gpus: int = 0,
    ):
        format_list = "\n".join(format_backend(BACKEND_LIST))
        assert backend in BACKEND_LIST, f"Choose backend in \n\n{format_list}"
        if not type(num_gpus) == int:
            raise TypeError(f"num_gpus should be of type int, not {type(num_gpus)}.")

        if "esm" in backend:
            model_dir = backend
            return TransformersWrapper(
                model_dir=model_dir, language_model_cls=ESMWrapper, num_gpus=num_gpus
            )
        else:
            model_dir = MAPPING_PROTBERT[backend]
            return TransformersWrapper(
                model_dir=model_dir,
                language_model_cls=RostlabWrapper,
                num_gpus=num_gpus,
            )

    @staticmethod
    def list_backend() -> None:
        """Get all possible backend for the model"""
        print(
            "Use backend in this list :\n\n",
            "\n".join(format_backend(BACKEND_LIST)),
            sep="",
        )



"""
This script defines a parent class for transformers, for which child classes which are
specific to a given transformers implementation can inherit.
It allows to derive probabilities, embeddings and log-likelihoods based on inputs
sequences, and displays some properties of the transformer model.
"""
import math
import os
import time
from copy import deepcopy
from os.path import join
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Type, Union

import numpy as np
import pytorch_lightning as pl
import ray
import torch
import torch.tensor
from biotransformers.utils.compute_utils import Mutation, get_list_probs, mutation_score
from biotransformers.utils.constant import NATURAL_AAS_LIST
from biotransformers.utils.logger import logger  # noqa
from biotransformers.utils.tqdm_utils import ProgressBar
from biotransformers.utils.utils import (
    get_logs_version,
    init_model_sequences,
    load_fasta,
)
from biotransformers.wrappers.language_model import LanguageModel
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import CSVLogger

from ..lightning_utils.data import create_dataloader
from ..lightning_utils.models import LightningModule

log = logger("transformers_wrapper")
PathMsaFolder = str
TokenProbsDict = Dict[int, Dict[str, float]]
SequenceProbsList = List[TokenProbsDict]


class TransformersWrapper:
    """
    Abstract class that uses pretrained transformers model to evaluate
    a protein likelihood so as other insights.
    """

    def __init__(
        self,
        model_dir: str,
        language_model_cls: Type[LanguageModel],
        num_gpus: int = 0,
    ):
        """Initialize Transformers wrapper

        TODO : os.environ["CUDA_VISIBLE_DEVICES"]="0" or export CUDA_VISIBLE_DEVICES="0,1"

        Args:
            model_dir: name directory of the pretrained model
            num_gpus: number of gpus to use. If set to 0, it uses the cpu.
        """
        self._model_dir = model_dir
        self._num_gpus = num_gpus
        self.language_model_cls = language_model_cls
        if num_gpus >= 1:
            assert torch.cuda.is_available(), "Cuda is not available."
            assert torch.cuda.device_count() >= num_gpus, "Not enough available GPUs."
        if num_gpus <= 1:
            device = "cpu" if num_gpus == 0 else "cuda"
            self._language_model = language_model_cls(
                model_dir=model_dir, device=device
            )
            self._multi_gpus = False
        else:
            self._language_model = language_model_cls(model_dir=model_dir, device="cpu")
            self._ray_cls = None
            self._workers = None
            self._multi_gpus = True

    def init_ray_workers(self):
        """Initialization of ray workers"""
        if self._multi_gpus:
            self._ray_cls = ray.remote(num_cpus=2, num_gpus=1)(self.language_model_cls)
            self._workers = [
                self._ray_cls.remote(model_dir=self._model_dir, device="cuda:0")
                for _ in range(self._num_gpus)
            ]

    def delete_ray_workers(self):
        """Delete ray workers to free RAM"""
        if self._multi_gpus:
            # kill worker to free RAM
            _ = [ray.kill(worker) for worker in self._workers]
            time.sleep(1)  # 1 sec timer to be sure ray workers are killed
            self._ray_cls = None
            self._workers = None

    def get_vocabulary_mask(self, tokens_list: List[str]) -> np.ndarray:
        """Returns a mask ove the model tokens."""
        # Compute a mask over vocabulary to decide which value to keep or not
        vocabulary_mask = np.array(
            [
                1.0 if token in tokens_list else 0.0
                for token in self._language_model.model_vocabulary
            ]
        )
        return vocabulary_mask

    def _get_num_batch_iter(self, model_inputs: Dict[str, Any], batch_size: int) -> int:
        """
        Get the number of batches when spliting model_inputs into chunks of size batch_size.
        """
        num_of_sequences = model_inputs["input_ids"].shape[0]
        num_batch_iter = int(np.ceil(num_of_sequences / batch_size))
        return num_batch_iter

    def _generate_chunks(
        self, model_inputs: Dict[str, Any], batch_size: int
    ) -> Generator[Dict[str, Iterable], None, None]:
        """Yield a dictionnary of tensor"""
        num_of_sequences = model_inputs["input_ids"].shape[0]
        for i in range(0, num_of_sequences, batch_size):
            batch_sequence = {
                key: value[i : (i + batch_size)] for key, value in model_inputs.items()
            }
            yield batch_sequence

    def _repeat_and_mask_inputs(
        self, model_inputs: Dict[str, torch.Tensor]
    ) -> Tuple[Dict[str, torch.Tensor], List[List]]:
        """Create new tensor by masking each token and repeating sequence

        Args:
            model_inputs: shape -> (num_seqs, max_seq_len)

        Returns:
            model_inputs: shape -> (sum_tokens, max_seq_len)
            masked_ids_list: len -> (num_seqs)
        """
        new_input_ids = []
        new_attention_mask = []
        new_token_type_ids = []
        mask_ids = []
        for sequence, binary_mask, zeros in zip(
            model_inputs["input_ids"],
            model_inputs["attention_mask"],
            model_inputs["token_type_ids"],
        ):
            mask_id = []
            mask_token = self._language_model.mask_token
            does_end_token_exist = self._language_model.does_end_token_exist
            for i in range(1, sum(binary_mask) - does_end_token_exist * 1):
                mask_sequence = torch.tensor(
                    sequence[:i].tolist()
                    + [self._language_model.token_to_id(mask_token)]
                    + sequence[i + 1 :].tolist(),
                    dtype=torch.int64,
                )
                new_input_ids.append(mask_sequence)
                new_attention_mask.append(binary_mask)
                new_token_type_ids.append(zeros)
                mask_id.append(i)
            mask_ids.append(mask_id)

        model_inputs_out = deepcopy(model_inputs)
        model_inputs_out["input_ids"] = torch.stack(new_input_ids)
        model_inputs_out["attention_mask"] = torch.stack(new_attention_mask)
        model_inputs_out["token_type_ids"] = torch.stack(new_token_type_ids)

        return model_inputs_out, mask_ids

    def _mask_inputs_tokens(
        self, model_inputs: Dict[str, torch.Tensor], token_position: Optional[List[int]]
    ) -> Dict[str, torch.Tensor]:
        """Create new tensor by masking a specific token

        Args:
            model_inputs (Dict[str, torch.Tensor]): [description]
            token_position (int): the position of the token to mask

        Returns:
            Tuple[Dict[str, torch.Tensor], List[List]]: [description]
        """
        if not isinstance(token_position, List):
            raise TypeError("masked_token_position should be of a list of integers.")
        if not isinstance(token_position[0], int):
            raise TypeError("index inside masked_token_position should be of type int.")
        for position in token_position:
            if position < 1:
                raise ValueError("Token indexation starts at 1.")

        new_input_ids = []
        new_attention_mask = []
        new_token_type_ids = []
        for sequence, binary_mask, zeros, position in zip(
            model_inputs["input_ids"],
            model_inputs["attention_mask"],
            model_inputs["token_type_ids"],
            token_position,
        ):
            mask_token = self._language_model.mask_token
            # Don't forget <CLS> token for index
            # position index start at 1
            mask_sequence = sequence.clone()
            mask_sequence[position] = self._language_model.token_to_id(mask_token)

            new_input_ids.append(mask_sequence)
            new_attention_mask.append(binary_mask)
            new_token_type_ids.append(zeros)

        model_inputs_out = deepcopy(model_inputs)
        model_inputs_out["input_ids"] = torch.stack(new_input_ids)
        model_inputs_out["attention_mask"] = torch.stack(new_attention_mask)
        model_inputs_out["token_type_ids"] = torch.stack(new_token_type_ids)

        return model_inputs_out

    def _gather_masked_outputs(
        self, model_outputs: torch.Tensor, masked_ids_list: List[List]
    ) -> torch.Tensor:
        """Gather all the masked outputs to get original tensor shape

        Args:
            model_outputs (torch.Tensor): shape -> (sum_tokens, max_seq_len, vocab_size)
            masked_ids_list (List[List]) : len -> (num_seqs)

        Returns:
            model_outputs (torch.Tensor): shape -> (num_seqs, max_seq_len, vocab_size)
        """
        max_length = model_outputs.shape[1]
        inf_tensor = -float("Inf") * torch.ones(
            [1, model_outputs.shape[2]], dtype=torch.float32
        )
        sequences_list = []
        start_id = 0
        for mask_id in masked_ids_list:
            end_id = start_id + len(mask_id)
            sequence = torch.cat(
                (
                    inf_tensor,
                    model_outputs[range(start_id, end_id), mask_id],
                    inf_tensor.repeat(max_length - len(mask_id) - 1, 1),
                ),
                0,
            )
            sequences_list.append(sequence)
            start_id = end_id
        return torch.stack(sequences_list)

    def _model_evaluation(
        self, model_inputs: Dict[str, torch.tensor], batch_size: int = 1, **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute logits and embeddings

        Function which computes logits and embeddings based on a list of sequences,
        a provided batch size and an inference configuration. The output is obtained
        by computing a forward pass through the model ("forward inference")

        Args:
            model_inputs (Dict[str, torch.tensor]): [description]
            batch_size (int): [description]

        Returns:
            Tuple[torch.tensor, torch.tensor]:
                    * logits [num_seqs, max_len_seqs, vocab_size]
                    * embeddings [num_seqs, max_len_seqs+1, embedding_size]
        """
        silent = kwargs.get("silent", False)
        num_inputs = model_inputs["input_ids"].shape[0]
        n_updates = math.ceil(num_inputs / batch_size)
        if self._multi_gpus:
            jobs = []
            # Initialize logits and embeddings before looping over batches
            logits = torch.Tensor()  # [num_seqs, max_len_seqs+1, vocab_size]
            embeddings = torch.Tensor()  # [num_seqs, max_len_seqs+1, embedding_size]
            pb = ProgressBar(n_updates)
            actor = pb.actor
            for i, batch_inputs in enumerate(
                self._generate_chunks(
                    model_inputs, math.ceil(num_inputs / self._num_gpus)
                )
            ):
                # Split large batch into smaller batches, when per GPU worker
                # Send tqdm progress bar
                jobs.append(
                    self._workers[i].model_pass.remote(  # type: ignore
                        batch_inputs, batch_size, silent, actor
                    )
                )
            pb.print_until_done()
            # Launch parallel execution in background
            outs = ray.get(jobs)
            # Gather the result
            for batch_logits, batch_embeddings in outs:
                embeddings = torch.cat((embeddings, batch_embeddings), dim=0)
                logits = torch.cat((logits, batch_logits), dim=0)
        else:
            logits, embeddings = self._language_model.model_pass(
                model_inputs, batch_size, silent
            )

        return logits, embeddings

    def _compute_logits(
        self,
        model_inputs: Dict[str, torch.Tensor],
        batch_size: int,
        pass_mode: str,
        **kwargs,
    ) -> torch.Tensor:
        """Intermediate function to compute logits

        Args:
            model_inputs[str] (torch.Tensor): shape -> (num_seqs, max_seq_len)
            batch_size (int)
            pass_mode (str)

        Returns:
            logits (torch.Tensor): shape -> (num_seqs, max_seq_len, vocab_size)
        """
        if pass_mode == "masked":
            if self._language_model.is_msa:
                raise NotImplementedError(
                    "Masked with msa-transformers is not implemented."
                )
            model_inputs, masked_ids_list = self._repeat_and_mask_inputs(model_inputs)
            logits, _ = self._model_evaluation(
                model_inputs, batch_size=batch_size, **kwargs
            )
            logits = self._gather_masked_outputs(logits, masked_ids_list)
        elif pass_mode == "forward":
            logits, _ = self._model_evaluation(
                model_inputs, batch_size=batch_size, **kwargs
            )
        return logits

    def compute_logits(
        self,
        sequences: Union[List[str], str],
        batch_size: int = 1,
        pass_mode: str = "forward",
        silent: bool = False,
        n_seqs_msa: int = 6,
    ) -> List[np.ndarray]:
        """Function that computes the logits from sequences.

        It returns a list of logits arrays for each sequence. If working with MSA,
        return a list of logits for each sequence of the MSA.

        Args:
            sequences : List of sequences, path of fasta file or path to a folder
                        with msa to a3m format.
            batch_size: number of sequences to consider for the forward pass
            pass_mode: Mode of model evaluation ('forward' or 'masked')
            silent: whether to print progress bar in console
            n_seqs_msa: number of sequence to consider in an msa file.
        Returns:
            List[np.ndarray]: logits in np.ndarray format
        """
        sequences, lengths = init_model_sequences(
            sequences=sequences,
            model_dir=self._model_dir,
            model_is_msa=self._language_model.is_msa,
            n_seqs_msa=n_seqs_msa,
            vocab_size=self._language_model.vocab_size,
            pass_mode=pass_mode,
            tokens_list=None,
        )

        self.init_ray_workers()
        # Perform inference in model to compute the logits
        inputs = self._language_model.process_sequences_and_tokens(sequences)
        labels = torch.unsqueeze(inputs["input_ids"], dim=-1)
        logits = self._compute_logits(inputs, batch_size, pass_mode, silent=silent)

        # Remove padded logits
        logits = [
            torch.from_numpy(logit.numpy().transpose()[:, 1 : (length + 1)].transpose())
            for logit, length in zip(list(logits), lengths)
        ]
        labels = [
            torch.from_numpy(label.numpy().transpose()[:, 1 : (length + 1)].transpose())
            for label, length in zip(list(labels), lengths)
        ]

        # Keep only corresponding to amino acids that are in the sequence
        logits = [
            torch.gather(logit, dim=-1, index=label).numpy()
            for logit, label in zip(logits, labels)
        ]
        # List of arrays of shape (seq_length, 1)
        self.delete_ray_workers()
        return logits

    def compute_probabilities(
        self,
        sequences: Union[List[str], str],
        batch_size: int = 1,
        tokens_list: List[str] = None,
        pass_mode: str = "forward",
        silent: bool = False,
        n_seqs_msa: int = 6,
        masked_token_position: Optional[List[int]] = None,
    ) -> Union[SequenceProbsList, List[SequenceProbsList]]:
        """Function that computes the probabilities over amino-acids from sequences.

        It takes as inputs a list of sequences and returns a list of dictionaries.
        Each dictionary contains the probabilities over the natural amino-acids for each
        position in the sequence. The keys represent the positions (indexed
        starting with 0) and the values are dictionaries of probabilities over
        the natural amino-acids for this position.

        When working with MSA, it returns a list of dictionnary for each sequence in the MSA.
        In these dictionaries, the keys are the amino-acids and the value
        the corresponding probabilities.

        Both ProtBert and ESM models have more tokens than the 20 natural amino-acids
        (for instance MASK or PAD tokens). It might not be of interest to take these
        tokens into account when computing probabilities or log-likelihood. By default
        we remove them and compute probabilities only over the 20 natural amino-acids.
        This behavior can be overridden through the tokens_list argument that enable
        the user to choose the tokens to consider when computing probabilities.

        Args:
            sequences : List of sequences, path of fasta file or path to a folder
            with msa to a3m format.
            batch_size: number of sequences to consider for the forward pass
            tokens_list: List of tokens to consider
            pass_mode: Mode of model evaluation ('forward' or 'masked')
            silent : display or not progress bar
            n_seqs_msa: number of sequence to consider in an msa file.
            masked_token_position: List of positions of a specific token to mask for each sequence.
                                   Index from 1 to N for sequence of length N. Number of index
                                   to mask should be equal to the number of sequences.
        Returns:
            List[Dict[int, Dict[str, float]]]: dictionaries of probabilities per seq
        """
        tokens_list = NATURAL_AAS_LIST if tokens_list is None else tokens_list
        sequences, lengths = init_model_sequences(
            sequences=sequences,
            model_dir=self._model_dir,
            model_is_msa=self._language_model.is_msa,
            n_seqs_msa=n_seqs_msa,
            vocab_size=self._language_model.vocab_size,
            pass_mode=pass_mode,
            tokens_list=tokens_list,
        )
        self.init_ray_workers()
        if pass_mode == "masked" and (masked_token_position is not None):
            raise ValueError(
                "Incompatible arguments. "
                "You can not specify a masked_token_position in masked mode."
            )

        # Perform inference in model to compute the logits
        inputs = self._language_model.process_sequences_and_tokens(sequences)
        if masked_token_position is not None:
            if len(masked_token_position) != len(sequences):
                raise ValueError(
                    "masked_token_position and sequences must have the same length."
                )
            inputs = self._mask_inputs_tokens(inputs, masked_token_position)
        logits = self._compute_logits(inputs, batch_size, pass_mode, silent=silent)
        # Remove padded logits
        # Use transpose so that function works for MSA and sequence
        logits = [
            torch.from_numpy(logit.numpy().transpose()[:, 1 : (length + 1)].transpose())
            for logit, length in zip(list(logits), lengths)
        ]
        # Set to -inf logits that correspond to tokens that are not in tokens list
        vocabulary_mask = torch.from_numpy(self.get_vocabulary_mask(tokens_list))
        # Avoid printing warnings
        np.seterr(divide="ignore")

        # Put -inf on character not in token list. Shape of this mask depends on

        masked_logits = []
        for logit in logits:
            # repeat MSA dimension
            if self._language_model.is_msa:
                repeat_dim = (logit.shape[0], logit.shape[1], 1)  # type: ignore
            else:
                repeat_dim = (logit.shape[0], 1)  # type: ignore
            masked_logit = logit + torch.from_numpy(
                np.tile(np.log(vocabulary_mask), repeat_dim)
            )
            masked_logits.append(masked_logit)
        # Use softmax to compute probabilities from logits
        # Due to the -inf, probs of tokens that are not in token list will be zero
        softmax = torch.nn.Softmax(dim=-1)
        probabilities = [softmax(logits) for logits in masked_logits]

        def _get_probabilities_dict(probs: torch.Tensor) -> Dict[str, float]:
            return {
                token: float(probs[i].cpu().numpy())
                for i, token in enumerate(self._language_model.model_vocabulary)
                if token in tokens_list
            }

        probabilities_dict_type = Union[SequenceProbsList, List[SequenceProbsList]]
        if self._language_model.is_msa:
            probabilities_dict: probabilities_dict_type = [
                [
                    {
                        key: _get_probabilities_dict(value)
                        for key, value in dict(enumerate(probs)).items()
                    }
                    for probs in msa
                ]
                for msa in probabilities
            ]
        else:
            probabilities_dict = [
                {
                    key: _get_probabilities_dict(value)
                    for key, value in dict(enumerate(probs)).items()
                }
                for probs in probabilities
            ]
        self.delete_ray_workers()
        return probabilities_dict  # type: ignore

    def compute_loglikelihood(
        self,
        sequences: Union[List[str], str],
        batch_size: int = 1,
        tokens_list: List[str] = None,
        pass_mode: str = "forward",
        silent: bool = False,
        normalize: bool = True,
        masked_token_position: Optional[List[int]] = None,
    ) -> List[float]:
        """Function that computes loglikelihoods of sequences.
        It returns a list of float values.

        Both ProtBert and ESM models have more tokens than the 20 natural amino-acids
        (for instance MASK or PAD tokens). It might not be of interest to take these
        tokens into account when computing probabilities or log-likelihood. By default
        we remove them and compute probabilities only over the 20 natural amino-acids.
        This behavior can be overridden through the tokens_list argument that enable
        the user to choose the tokens to consider when computing probabilities.
        By default, all loglikelihoods are normalized by the sequence length. For
        example, a loglikelihood of -0.35 means that every amino acids are predicted with
        a probability of 0.7 in average.

        Args:
            sequences: List of sequences
            batch_size: Batch size
            tokens_list: List of tokens to consider
            pass_mode: Mode of model evaluation ('forward' or 'masked')
            silent : display or not progress bar
            normalize : If True, loglikelihood are normalize by sequence length.
            masked_token_position: List of positions of a specific token to mask for each sequence.
                                   Index from 1 to N for sequence of length N. Number of index to mask
                                   should be equal to the number of sequences.

        Returns:
            List[float]: list of loglikelihoods, one per sequence
        """
        tokens_list = NATURAL_AAS_LIST if tokens_list is None else tokens_list
        sequences, lengths = init_model_sequences(
            sequences=sequences,
            model_dir=self._model_dir,
            model_is_msa=self._language_model.is_msa,
            n_seqs_msa=0,
            vocab_size=self._language_model.vocab_size,
            pass_mode="forward",
            tokens_list=tokens_list,
        )
        self.init_ray_workers()
        if self._language_model.is_msa:
            raise NotImplementedError(
                "compute_loglikelihood for MSA transformers is not implemented."
            )

        probabilities = self.compute_probabilities(
            sequences,
            batch_size,
            tokens_list,
            pass_mode,
            silent,
            masked_token_position=masked_token_position,
        )
        log_likelihoods = []
        for sequence, probabilities_dict in zip(sequences, probabilities):
            log_likelihood = np.sum(
                [np.log(probabilities_dict[i][sequence[i]]) for i in range(len(sequence))]  # type: ignore
            )
            log_likelihoods.append(float(log_likelihood))
        if normalize:
            log_likelihoods = [
                log / length for log, length in zip(log_likelihoods, lengths)
            ]

        self.delete_ray_workers()
        return log_likelihoods

    def compute_mutation_score(
        self,
        sequences: Union[List[str], str],
        mutations: List[List[str]],
        batch_size: int = 1,
        tokens_list: List[str] = None,
        silent: bool = False,
    ) -> List[float]:
        """Function that computes loglikelihoods of sequences.
        It returns a list of float values.

        This function is used to score the a mutation between two amino acids as described in
        https://www.biorxiv.org/content/10.1101/2021.07.09.450648v1.full.pdf. This metrics is
        maximizedin ESM-1V to assess the interest of a mutation. The mutational score is based on
        the masked marginal probability (L forward passes) where whe introduce a mask at the
        mutation position and compute the log difference of probability between native and
        mutate sequence.

        score -> Sum(log(p(xi=xi_mutate|x-M))-log(p(xi=xi_native|x-M))) over M (M s a mutation set)

        The function takes in input a list of mutations for each sequence to evaluate.
        Mutations are tuple a single mutation, you can provide multiple mutations for a single sequence.
        Example:
            mutation: "A8E" to mutate amino acids A by E at position 8.
            mutation are indexed from 1 to N for sequence of length N.

            Below a mutations list for 3 sequences. We have to provide 3 tuples of single mutation.
            mutations: [["A3U","A8E"],["B7I"],["I124","E1J"]]

        Args:
            sequences: List of sequences
            batch_size: Batch size
            tokens_list: List of tokens to consider
            silent : display or not progress bar
            mutations: List of mutations for each sequence to evaluate. Mutations are list a single mutation.
                       mutation are indexed from 1 to N for sequence of length N.
        Returns:
            List[float]: list of mutations score for each sequence
        """
        tokens_list = NATURAL_AAS_LIST if tokens_list is None else tokens_list
        sequences, _ = init_model_sequences(
            sequences=sequences,
            model_dir=self._model_dir,
            model_is_msa=self._language_model.is_msa,
            n_seqs_msa=0,
            vocab_size=self._language_model.vocab_size,
            pass_mode=None,
            tokens_list=tokens_list,
        )
        self.init_ray_workers()
        if self._language_model.is_msa:
            raise NotImplementedError(
                "compute_mutation_score for MSA transformers is not implemented."
            )
        if len(mutations) != len(sequences):
            raise ValueError("mutation_list and sequences must have the same length.")

        mutations_list = [tuple((Mutation(mut) for mut in tup)) for tup in mutations]
        new_sequence = []
        mutations_index = []
        # expand sequence based on number of mutations an check mutation
        for seq, mutation in zip(sequences, mutations_list):
            for single_mutation in mutation:
                single_mutation.is_valid_mutation(seq)
                mutations_index.append(single_mutation.position)
                new_sequence.append(seq)

        # compute probas for all mutant in batch
        mutate_probabilities = self.compute_probabilities(
            new_sequence,
            batch_size,
            tokens_list,
            "forward",
            silent=silent,
            masked_token_position=mutations_index,
        )
        length_mutations = [
            len(mut) for mut in mutations_list
        ]  # use for reshape all mutations
        native_probs, mutate_probs = get_list_probs(mutations_list, mutate_probabilities, length_mutations)  # type: ignore
        mutation_score_list = [
            mutation_score(n_probs, m_probs)
            for n_probs, m_probs in zip(native_probs, mutate_probs)
        ]
        self.delete_ray_workers()
        return mutation_score_list

    def compute_embeddings(
        self,
        sequences: Union[List[str], str],
        batch_size: int = 1,
        pool_mode: Tuple[str, ...] = ("cls", "mean", "full"),
        silent: bool = False,
        n_seqs_msa: int = 6,
    ) -> Dict[str, Union[List[np.ndarray], np.ndarray]]:
        """Function that computes embeddings of sequences.

        The embedding of one sequence has a shape (sequence_length, embedding_size)
        where embedding_size equals 768 or 1024., thus we may want to use an aggregation
        function specified in pool_mode to aggregate the tensor on the num_tokens dimension.
        It might for instance avoid blowing the machine RAM when computing embeddings
        for a large number of sequences.

        'mean' signifies that we take the mean over the num_tokens dimension. 'cls'
        means that only the class token embedding is used.

        This function returns a dictionary of lists. The dictionary will have one key
        per pool-mode that has been specified. The corresponding value is a list of
        embeddings, one per sequence in sequences.

        When working with MSA, an extra dimension is added to the final tensor.
        Args:
            sequences: List of sequences, path of fasta file or path to a folder with msa to a3m format.
            batch_size: batch size
            pool_mode: Mode of pooling ('cls', 'mean', 'full')
            silent : whereas to display or not progress bar
            n_seqs_msa: number of sequence to consider in an msa file.
        Returns:
             Dict[str, List[np.ndarray]]: dict matching pool-mode and list of embeddings
        """
        sequences, lengths = init_model_sequences(
            sequences=sequences,
            model_dir=self._model_dir,
            model_is_msa=self._language_model.is_msa,
            n_seqs_msa=n_seqs_msa,
            vocab_size=self._language_model.vocab_size,
            pool_mode=pool_mode,
        )
        self.init_ray_workers()
        # Compute a forward pass to get the embeddings
        inputs = self._language_model.process_sequences_and_tokens(sequences)
        _, embeddings = self._model_evaluation(
            inputs, batch_size=batch_size, silent=silent
        )
        embeddings = [emb.cpu().numpy() for emb in embeddings]
        # Remove class token and padding
        # Use tranpose to filter on the two last dimensions. Doing this, we don't have to manage
        # the first dimension of the tensor. It works for [dim1, dim2, token_size, emb_size] and
        # for [dim1, token_size, emb_size]
        filtered_embeddings = [
            emb.transpose()[:, 1 : (length + 1)].transpose()
            for emb, length in zip(list(embeddings), lengths)
        ]
        # Keep class token only
        cls_embeddings = [emb.transpose()[:, 0].transpose() for emb in list(embeddings)]
        embeddings_dict = {}
        # Keep only what's necessary
        if "full" in pool_mode:
            embeddings_dict["full"] = filtered_embeddings
        if "cls" in pool_mode:
            embeddings_dict["cls"] = np.stack(cls_embeddings)
        if "mean" in pool_mode:
            # For msa embbedings, we average over tokens of each msa, we don't average over sequence.
            # We use transpose and average over axis 1 to not take in account msa dimension
            # esm model: [tokens , embedding] msa: [n_msa, tokens, embedding]
            embeddings_dict["mean"] = np.stack(
                [e.transpose().mean(1).transpose() for e in filtered_embeddings]
            )
        self.delete_ray_workers()
        return embeddings_dict

    def compute_accuracy(
        self,
        sequences: Union[List[str], str],
        batch_size: int = 1,
        pass_mode: str = "forward",
        silent: bool = False,
        n_seqs_msa: int = 6,
    ) -> float:
        """Compute model accuracy from the input sequences

        When working with MSA, the accuracy is computed over all the tokens of the msa' sequences.
        Args:
            sequences (Union[List[str],str]): List of sequences, path of fasta file or path to a folder with msa to a3m format.
            batch_size ([type], optional): [description]. Defaults to 1.
            pass_mode ([type], optional): [description]. Defaults to "forward".
            silent : whereas to display or not progress bar
            n_seqs_msa: number of sequence to consider in an msa file.
        Returns:
            float: model's accuracy over the given sequences
        """
        sequences, lengths = init_model_sequences(
            sequences=sequences,
            model_dir=self._model_dir,
            model_is_msa=self._language_model.is_msa,
            n_seqs_msa=n_seqs_msa,
            vocab_size=self._language_model.vocab_size,
            pass_mode=pass_mode,
        )
        self.init_ray_workers()

        # Perform inference in model to compute the logits
        inputs = self._language_model.process_sequences_and_tokens(sequences)
        logits = self._compute_logits(inputs, batch_size, pass_mode, silent=silent)
        # Get length of sequence
        labels = torch.unsqueeze(inputs["input_ids"], dim=-1)

        # Remove padded logits
        logits = [
            torch.from_numpy(logit.numpy().transpose()[:, :length].transpose())
            for logit, length in zip(list(logits), lengths)
        ]
        labels = [
            torch.from_numpy(label.numpy().transpose()[:, :length].transpose())
            for label, length in zip(list(labels), lengths)
        ]

        # Get the predicted labels
        predicted_labels = [torch.argmax(logit, dim=-1) for logit in logits]
        # Compute the accuracy
        accuracy = float(
            torch.mean(
                torch.cat(
                    [
                        torch.eq(predicted_label, label.squeeze()).float()
                        for predicted_label, label in zip(predicted_labels, labels)
                    ],
                    dim=0,
                ).float()
            )
        )
        self.delete_ray_workers()
        return accuracy

    def load_model(self, model_dir: str, map_location=None):
        """Load state_dict a finetune pytorch model ro a checkpoint directory

        More informations about how to load a model with map_location:
            https://pytorch.org/tutorials/beginner/saving_loading_models.html#saving-loading-model-for-inference

        Args:
            model_dir: path file of the pt model or checkpoint.
                       the checkpoint should be a pytorch model checkpoint
        """
        if not os.path.isfile(model_dir):
            raise FileNotFoundError

        if self._multi_gpus:
            self.init_ray_workers()
            ray.get(
                [worker._load_model.remote(model_dir, map_location) for worker in self._workers]  # type: ignore
            )
            pass
        else:
            self._language_model._load_model(model_dir, map_location)  # type: ignore

    def _save_model(self, exp_path: str, lightning_model: pl.LightningModule):
        """Save pytorch model in pytorch-lightning logs directory
        Args:
            exp_path (str): path of the experiments directory in the logs
        """
        version = get_logs_version(exp_path)
        model_dir = self._model_dir.replace("/", "_")
        if version is not None:
            save_name = os.path.join(exp_path, version, model_dir + "_finetuned.pt")
        else:
            save_name = os.path.join(exp_path, model_dir + "_finetuned.pt")
        torch.save(lightning_model.model.state_dict(), save_name)
        log.info("Model save at %s." % save_name)

        return save_name

    def finetune(
        self,
        train_sequences: Union[List[str], str],
        validation_sequences: Union[List[str], str],
        num_data_workers: int = 4,
        lr: float = 1.0e-5,
        warmup_updates: int = 1024,
        warmup_init_lr: float = 1e-7,
        epochs: int = 10,
        acc_batch_size: int = 50,
        masking_ratio: float = 0.025,
        masking_prob: float = 0.8,
        random_token_prob: float = 0.15,
        toks_per_batch: int = 2048,
        crop_sizes: Tuple[int, int] = (512, 1024),
        accelerator: str = "ddp",
        amp_level: str = "O2",
        precision: int = 16,
        logs_save_dir: str = "logs",
        logs_name_exp: str = "finetune_masked",
        checkpoint: Optional[str] = None,
        save_last_checkpoint: bool = True,
    ):
        """Function to finetune a model on a specific dataset

        This function will finetune the choosen model on a dataset of
        sequences with pytorch ligthening. You can modify the masking ratio of AA
        in the arguments for better convergence.
        Be careful with the accelerator that you use. DDP accelerator will
        launch multiple python process and do not be use in a notebook.

        More informations on GPU/accelerator compatibility here :
            https://pytorch-lightning.readthedocs.io/en/stable/advanced/multi_gpu.html
        The wisest choice would be to use DDP for multi-gpu training.

        Args:
            train_sequences : Could be a list of sequences or the path of a
                              fasta file with multiple seqRecords
            validation_sequences : Could be a list of sequences or the path of a
                              fasta file with multiple seqRecords
            num_data_workers: number of cpus workers per gpu to load data
            lr : learning rate for training phase. Defaults to 1.0e-5.
            warmup_updates : Number of warming updates, number of step while increasing
            the leraning rate. Defaults to 1024.
            warmup_init_lr :  Initial lr for warming_update. Defaults to 1e-7.
            epochs :  number of epoch for training. Defaults to 10.
            acc_batch_size : number of batches of toks_per_batch tokens to accumulate before applying gradients.
            masking_ratio : ratio of tokens to be masked. Defaults to 0.025.
            masking_prob :  probability that the chose token is replaced with a mask token.
                            Defaults to 0.8.
            random_token_prob : probability that the chose token is replaced with a random token.
                                Defaults to 0.1.
            toks_per_batch: Maximum number of token to consider in a batch.Defaults to 2048.
                            This argument will set the number of sequences in a batch, which
                            is dynamically computed. Batch size use accumulate_grad_batches
                            to compute accumulate_grad_batches parameter.
            crop_sizes: range of lengths to crop dynamically sequences when sampling them
            extra_toks_per_seq: Defaults to 2,
            accelerator: type of accelerator for mutli-gpu processing (DPP recommanded)
            amp_level: allow mixed precision. Defaults to '02'
            precision: reducing precision allows to decrease the GPU memory needed.
                       Defaults to 16 (float16)
            logs_save_dir : Defaults directory to logs.
            logs_name_exp: Name of the experience in the logs.
            checkpoint : Path to a checkpoint file to restore training session.
            save_last_checkpoint: Save last checkpoint and 2 best trainings models
                                  to restore training session. Take a large amout of time
                                  and memory.
        """
        if isinstance(train_sequences, str):
            train_sequences = load_fasta(train_sequences)

        if isinstance(validation_sequences, str):
            validation_sequences = load_fasta(validation_sequences)

        if self._language_model.is_msa:
            raise NotImplementedError("MSA finetunening not implemented.")

        fit_model = self._language_model.model  # type: ignore
        alphabet = self._language_model.get_alphabet_dataloader()

        if alphabet.model_dir == "esm1b_t33_650M_UR50S":
            if crop_sizes[1] > 1024:
                raise ValueError(
                    "ESM-1b model works only with sequences of maximum "
                    "length 1024. Please, be sure to crop to maximum of "
                    "1024 when using this model."
                )

        lightning_model = LightningModule(
            model=fit_model,
            alphabet=alphabet,
            lr=lr,
            warmup_updates=warmup_updates,
            warmup_init_lr=warmup_init_lr,
            warmup_end_lr=lr,
        )

        train_dataloader = create_dataloader(
            sequences=train_sequences,
            alphabet=alphabet,
            masking_ratio=masking_ratio,
            masking_prob=masking_prob,
            random_token_prob=random_token_prob,
            num_workers=num_data_workers,
            toks_per_batch=toks_per_batch,
            crop_sizes=crop_sizes,
        )

        validation_dataloader = create_dataloader(
            sequences=validation_sequences,
            alphabet=alphabet,
            masking_ratio=masking_ratio,
            masking_prob=masking_prob,
            random_token_prob=random_token_prob,
            num_workers=num_data_workers,
            toks_per_batch=toks_per_batch,
            crop_sizes=crop_sizes,
        )

        if self._num_gpus == 0:
            pass
            # raise ValueError("You try to train a transformers without GPU.")

        logger = CSVLogger(logs_save_dir, name=logs_name_exp)
        checkpoint_callback = None

        if save_last_checkpoint:
            checkpoint_callback = [
                ModelCheckpoint(
                    save_last=True,
                    save_top_k=2,
                    mode="max",
                    monitor="val_acc",
                    every_n_val_epochs=3,
                )
            ]

        trainer = Trainer(
            gpus=self._num_gpus,
            amp_level=amp_level,
            precision=precision,
            accumulate_grad_batches=acc_batch_size,
            max_epochs=epochs,
            logger=logger,
            accelerator=accelerator,
            replace_sampler_ddp=False,
            resume_from_checkpoint=checkpoint,
            callbacks=checkpoint_callback,
        )
        trainer.fit(lightning_model, train_dataloader, [validation_dataloader])

        save_path = str(Path(join(logs_save_dir, logs_name_exp)).resolve())
        if accelerator == "ddp":
            rank = os.environ.get("LOCAL_RANK", None)
            rank = int(rank) if rank is not None else None  # type: ignore
            if rank == 0:
                save_name = self._save_model(save_path, lightning_model)
        else:
            save_name = self._save_model(save_path, lightning_model)

        # Load new model
        self._language_model._load_model(save_name)
        log.info("Training completed.")



"""Test module for testing accuracy function"""
import pytest

test_params = [
    (1, "forward"),
    (2, "masked"),
    (10, "forward"),
]

test_params_fasta = [(2, "forward")]


@pytest.mark.parametrize("batch_size, pass_mode", test_params)
def test_accuracy_type_and_range(init_model, sequences, batch_size, pass_mode):
    test_trans = init_model
    accuracy = test_trans.compute_accuracy(
        sequences,
        batch_size=batch_size,
        pass_mode=pass_mode,
    )
    assert isinstance(accuracy, float)
    assert (accuracy >= 0.0) and (accuracy <= 1.0)


@pytest.mark.parametrize("batch_size, pass_mode", test_params_fasta)
def test_accuracy_type_and_range_fasta(init_model, fasta_path, batch_size, pass_mode):
    test_trans = init_model

    accuracy_fasta = test_trans.compute_accuracy(
        fasta_path,
        batch_size=batch_size,
        pass_mode=pass_mode,
    )
    assert isinstance(accuracy_fasta, float)
    assert (accuracy_fasta >= 0.0) and (accuracy_fasta <= 1.0)



"""Test module for testing loglikelihoods function"""
import pytest
from numpy.testing import assert_allclose

test_params = [
    (1, list("ACDEFGHIKLMNPQRSTVWY"), "forward", "params1"),
    (2, list("ACDEFGHIKLMNPQRSTVWY") + ["MASK"], "masked", "params2"),
    (10, list("ACDEFGHIKLMNPQRSTVWY") + ["MASK"], "forward", "params3"),
]

test_params_fasta = [(1, list("KFQRVACEXWIHYPNGSMTDL"), "forward")]


@pytest.mark.parametrize("batch_size, tokens_list, pass_mode, params", test_params)
def test_loglikelihoods_type_shape_and_range(
    init_model,
    sequences,
    loglikelihoods_results,
    batch_size,
    tokens_list,
    pass_mode,
    params,
):
    test_trans = init_model
    loglikelihoods = test_trans.compute_loglikelihood(
        sequences,
        batch_size=batch_size,
        tokens_list=tokens_list,
        pass_mode=pass_mode,
        normalize=True,
    )
    assert len(loglikelihoods) == len(sequences)
    if test_trans._model_dir in loglikelihoods_results.keys():
        results = loglikelihoods_results[test_trans._model_dir][params]
        assert_allclose(loglikelihoods, results, rtol=0.01)


@pytest.mark.parametrize("batch_size, tokens_list, pass_mode", test_params_fasta)
def test_loglikelihoods_type_shape_and_range_fasta(
    init_model,
    fasta_path,
    lengths_sequence_fasta,
    loglikelihoods_fasta_results,
    batch_size,
    tokens_list,
    pass_mode,
):
    test_trans = init_model
    loglikelihoods = test_trans.compute_loglikelihood(
        fasta_path,
        batch_size=batch_size,
        tokens_list=tokens_list,
        pass_mode=pass_mode,
        normalize=True,
    )
    assert len(loglikelihoods) == len(lengths_sequence_fasta)
    if test_trans._model_dir in loglikelihoods_fasta_results.keys():
        results = loglikelihoods_fasta_results[test_trans._model_dir]
        assert_allclose(loglikelihoods, results, rtol=0.01)



"""
This script defines a class which inherits from the LanguageModel class, and is
specific to the Rostlab models (eg ProtBert and ProtBert-BFD) developed by
hugging face
- ProtBert: https://huggingface.co/Rostlab/prot_bert
- ProtBert BFD: https://huggingface.co/Rostlab/prot_bert_bfd
"""
from typing import Dict, List, Tuple

import torch
import copy
from biotransformers.lightning_utils.data import (
    AlphabetDataLoader,
    convert_ckpt_to_statedict,
)
from biotransformers.utils.constant import DEFAULT_ROSTLAB_MODEL, ROSTLAB_LIST
from biotransformers.utils.logger import logger  # noqa
from biotransformers.utils.utils import _generate_chunks, _get_num_batch_iter
from biotransformers.wrappers.language_model import LanguageModel
from ray.actor import ActorHandle
from tqdm import tqdm
from transformers import BertForMaskedLM, BertTokenizer

log = logger("rostlab_wrapper")


class RostlabWrapper(LanguageModel):
    """
    Class that uses a rostlab type of pretrained transformers model to evaluate
    a protein likelihood so as other insights.
    """

    def __init__(self, model_dir: str, device):
        if model_dir not in ROSTLAB_LIST:
            print(
                f"Model dir '{model_dir}' not recognized."
                f" Using '{DEFAULT_ROSTLAB_MODEL}' as default"
            )
            model_dir = DEFAULT_ROSTLAB_MODEL
        super().__init__(model_dir=model_dir, device=device)
        self.tokenizer = BertTokenizer.from_pretrained(
            model_dir, do_lower_case=False, padding=True
        )
        self._model = (
            BertForMaskedLM.from_pretrained(self._model_dir).eval().to(self._device)
        )
        self.hidden_size = self._model.config.hidden_size
        self.mask_pipeline = None

    @property
    def model(self) -> torch.nn.Module:
        """Return torch model."""
        return self._model

    @property
    def clean_model_id(self) -> str:
        """Clean model ID (in case the model directory is not)"""
        return self.model_id.replace("rostlab/", "")

    @property
    def model_vocabulary(self) -> List[str]:
        """Returns the whole vocabulary list"""
        return list(self.tokenizer.vocab.keys())

    @property
    def vocab_size(self) -> int:
        """Returns the whole vocabulary size"""
        return self.tokenizer.vocab_size

    @property
    def mask_token(self) -> str:
        """Representation of the mask token (as a string)"""
        return self.tokenizer.mask_token  # "[MASK]"

    @property
    def pad_token(self) -> str:
        """Representation of the pad token (as a string)"""
        return self.tokenizer.pad_token  # "[PAD]"

    @property
    def begin_token(self) -> str:
        """Representation of the beginning of sentence token (as a string)"""
        return self.tokenizer.cls_token  # "[CLS]"

    @property
    def end_token(self) -> str:
        """Representation of the end of sentence token (as a string)."""
        return self.tokenizer.sep_token  # "[SEP]"

    @property
    def does_end_token_exist(self) -> bool:
        """Returns true if a end of sequence token exists"""
        return True

    @property
    def token_to_id(self):
        """Returns a function which maps tokens to IDs"""
        return lambda x: self.tokenizer.convert_tokens_to_ids(x)

    @property
    def embeddings_size(self) -> int:
        """Returns size of the embeddings"""
        return self.hidden_size

    def _load_model(self, path_model: str, map_location=None):
        """Load model."""
        if path_model.endswith(".pt"):
            loaded_model = torch.load(path_model)
        elif path_model.endswith(".ckpt"):
            loaded_model = convert_ckpt_to_statedict(
                torch.load(path_model)["state_dict"]
            )
        else:
            raise ValueError("Expecting a .pt or .ckpt file")
        self._model.load_state_dict(loaded_model, map_location)
        self._model.eval()
        log.info("Load model %s" % path_model)

    def process_sequences_and_tokens(
        self,
        sequences_list: List[str],
    ) -> Dict[str, torch.tensor]:
        """Function to transform tokens string to IDs; it depends on the model used"""
        separated_sequences_list = [" ".join(seq) for seq in sequences_list]
        encoded_inputs = self.tokenizer(
            separated_sequences_list,
            return_tensors="pt",
            padding=True,
        ).to("cpu")
        return encoded_inputs

    def model_pass(
        self,
        model_inputs: Dict[str, torch.tensor],
        batch_size: int,
        silent: bool = False,
        pba: ActorHandle = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Function which computes logits and embeddings based on a dict of sequences
        tensors, a provided batch size and an inference configuration. The output is
        obtained by computing a forward pass through the model ("forward inference")

        Args:
            model_inputs (Dict[str, torch.tensor]): [description]
            batch_size (int): size of the batch
            silent : display or not progress bar
            pba : tqdm progress bar for ray actor
        Returns:
            Tuple[torch.tensor, torch.tensor]:
                    * logits [num_seqs, max_len_seqs, vocab_size]
                    * embeddings [num_seqs, max_len_seqs+1, embedding_size]
        """
        if pba is None:
            batch_generator = tqdm(
                _generate_chunks(model_inputs, batch_size),
                total=_get_num_batch_iter(model_inputs, batch_size),
                disable=silent,
            )
        else:
            batch_generator = _generate_chunks(model_inputs, batch_size)

        logits = torch.Tensor()  # [num_seqs, max_len_seqs+1, vocab_size]
        embeddings = torch.Tensor()  # [num_seqs, max_len_seqs+1, embedding_size]
        for batch_inputs in batch_generator:
            with torch.no_grad():
                model_inputs = {
                    key: value.to(self._device) for key, value in batch_inputs.items()
                }
                model_outputs = self._model(**model_inputs, output_hidden_states=True)
                batch_logits = model_outputs.logits.detach().cpu()
                batch_embeddings = model_outputs.hidden_states[-1].detach().cpu()
                embeddings = torch.cat((embeddings, batch_embeddings), dim=0)
                logits = torch.cat((logits, batch_logits), dim=0)

            # tqdm worker update
            if pba is not None:
                pba.update.remote(1)
        return logits, embeddings

    def get_alphabet_dataloader(self):
        """Define an alphabet mapping for common method between
        protbert and ESM
        """

        def tokenize(x: List[str]):
            x_ = [" ".join(seq) for seq in x]
            tokens = self.tokenizer(x_, return_tensors="pt", padding=True)
            return x, tokens["input_ids"]

        all_tokens = copy.deepcopy(self.tokenizer.vocab)
        del all_tokens["[PAD]"]
        del all_tokens["[UNK]"]
        del all_tokens["[CLS]"]
        del all_tokens["[SEP]"]
        del all_tokens["[MASK]"]
        standard_tokens = list(all_tokens.keys())

        alphabet_dl = AlphabetDataLoader(
            prepend_bos=True,
            append_eos=True,
            mask_idx=self.tokenizer.mask_token_id,
            pad_idx=self.tokenizer.pad_token_id,
            standard_toks=standard_tokens,
            model_dir=self._model_dir,
            lambda_toks_to_ids=lambda x: self.tokenizer.convert_tokens_to_ids(x),
            lambda_tokenizer=lambda x: tokenize(x),
        )
        return alphabet_dl



"""
This script defines a generic template class for any language model.
Both ESM and Rostlab language models should implement this class.
"""
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

import torch
from ray.actor import ActorHandle


class LanguageModel(ABC):
    """
    Class that implements a language model.
    """

    def __init__(self, model_dir: str, device):
        self._model_dir = model_dir
        self._device = device
        self.is_msa = False

    @property
    def model_id(self) -> str:
        """Model ID, as specified in the model directory"""
        return self._model_dir.lower()

    @property
    @abstractmethod
    def clean_model_id(self) -> str:
        """Clean model ID (in case the model directory is not)"""
        pass

    @property
    @abstractmethod
    def model_vocabulary(self) -> List[str]:
        """Returns the whole vocabulary list"""
        pass

    @property
    @abstractmethod
    def vocab_size(self) -> int:
        """Returns the whole vocabulary size"""
        pass

    @property
    @abstractmethod
    def mask_token(self) -> str:
        """Representation of the mask token (as a string)"""
        pass

    @property
    @abstractmethod
    def pad_token(self) -> str:
        """Representation of the pad token (as a string)"""
        pass

    @property
    @abstractmethod
    def begin_token(self) -> str:
        """Representation of the beginning of sentence token (as a string)"""
        pass

    @property
    @abstractmethod
    def end_token(self) -> str:
        """Representation of the end of sentence token (as a string)."""
        pass

    @property
    @abstractmethod
    def does_end_token_exist(self) -> bool:
        """Returns true if a end of sequence token exists"""
        pass

    @property
    @abstractmethod
    def token_to_id(self):
        """Returns a function which maps tokens to IDs"""
        pass

    @property
    @abstractmethod
    def embeddings_size(self) -> int:
        """Returns size of the embeddings"""
        pass

    @abstractmethod
    def process_sequences_and_tokens(
        self,
        sequences_list: List[str],
    ) -> Dict[str, torch.Tensor]:
        """Function to transform tokens string to IDs; it depends on the model used"""
        pass

    @property
    @abstractmethod
    def model(self) -> torch.nn.Module:
        """Return torch model."""
        pass

    @abstractmethod
    def _load_model(self, path: str):
        """Load model."""
        pass

    @abstractmethod
    def model_pass(
        self,
        model_inputs: Dict[str, torch.tensor],
        batch_size: int,
        silent: bool = False,
        pba: ActorHandle = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Function which computes logits and embeddings based on a dict of sequences
        tensors, a provided batch size and an inference configuration. The output is
        obtained by computing a forward pass through the model ("forward inference")

        Args:
            model_inputs (Dict[str, torch.tensor]): [description]
            batch_size (int): size of the batch
            silent : display or not progress bar
            pba : tqdm progress bar for ray actor
        Returns:
            Tuple[torch.tensor, torch.tensor]:
                    * logits [num_seqs, max_len_seqs, vocab_size]
                    * embeddings [num_seqs, max_len_seqs+1, embedding_size]
        """
        pass

    @abstractmethod
    def get_alphabet_dataloader(self):
        """Define an alphabet mapping for common method between
        protbert and ESM
        """
        pass



import math
import os
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Union

import numpy as np
from Bio import SeqIO
from biotransformers.utils.constant import BACKEND_LIST
from biotransformers.utils.logger import logger
from biotransformers.utils.msa_utils import get_msa_lengths, get_msa_list, read_msa

log = logger("utils")


def init_model_sequences(
    sequences: Union[List[str], str],
    model_dir: str,
    model_is_msa: bool,
    n_seqs_msa: int,
    vocab_size: int = 0,
    embeddings_size: int = 0,
    pass_mode: Optional[str] = None,
    pool_mode: Optional[Tuple[str, ...]] = None,
    tokens_list: Optional[List[str]] = None,
) -> Tuple[List, List]:
    """Function use at the beginning of each compute_function that load
    sequences depending on the type of model.
    When using MSA-model, return a list of MSA, which is simply a list of sequences with special tokens.

    Args:
        sequences (Union[List[str], str]): List of sequences, path of fasta file or path to a folder with msa to a3m format.
        model_dir (str): name of the model
        model_is_msa (bool): is MSA model
        n_seqs_msa (int): number of seqs to consider in an MSA
        vocab_size (Optional[int], optional): Defaults to 0.
        embeddings_size (Optional[int], optional): Defaults to 0.
        pass_mode (Optional[str], optional): "masked" or "forward". Defaults to None.
        pool_mode (Optional[Tuple[str, ...]], optional): full, mean or cls. Defaults to None.
        tokens_list (Optional[List[str]], optional): list of valid tokens. Defaults to None.

    Raises:
        ValueError: [description]

    Returns:
        Tuple[List,List]: return a list of sequence string/MSA and list of lenghts for each sequence/MSA.
    """
    if not model_is_msa:
        if isinstance(sequences, str):
            sequences = load_fasta(sequences)
        _check_sequence(sequences, model_dir, 1024)
        if pass_mode is not None:
            _check_memory_logits(sequences, vocab_size, pass_mode)
        if pool_mode is not None:
            _check_memory_embeddings(sequences, embeddings_size, pool_mode)

        if tokens_list is not None:
            _check_tokens_list(sequences, tokens_list)
        lengths = [len(sequence) for sequence in sequences]
    else:
        if not isinstance(sequences, str):
            raise ValueError("The path to MSA folder must be a string.")
        path_msa = str(Path(sequences).resolve())
        list_msa_filepath = get_msa_list(path_msa)
        sequences = [read_msa(file, n_seqs_msa) for file in list_msa_filepath]  # type: ignore
        lengths = get_msa_lengths(sequences, n_seqs_msa)  # type: ignore

    return sequences, lengths


def convert_bytes_size(size_bytes: int) -> Tuple[str, bool]:
    """[summary]

    Args:
        size_bytes: size in bytes

    Returns:
        Tuple[str,bool]: return the size with correct units and a condition
        to display the warning message.
    """
    if size_bytes == 0:
        return "0B", False
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = int(round(size_bytes / p, 2))
    is_warning = i >= 3  # warning on size only for model in GB

    return "%s%s" % (s, size_name[i]), is_warning


def _check_memory_embeddings(
    sequences_list: List[str], embeddings_size: int, pool_mode: Tuple[str, ...]
):
    """Function to compute the memory taken by the embeddings with float64 number.

    Args:
        sequences_list: sequences of proteins
        embeddings_size : size of the embeddings vector, depends on the model
        pool_mode : aggregation function
    """
    num_of_sequences = len(sequences_list)
    emb_dict_len = len(pool_mode)
    tensor_memory_bits = 64  # double/float64
    memory_bits = num_of_sequences * embeddings_size * emb_dict_len * tensor_memory_bits
    memory_bytes = int(memory_bits / 8)
    memory_convert_bytes, is_warning = convert_bytes_size(memory_bytes)

    if is_warning:
        log.warning(
            "Embeddings will need about %s of memory."
            "Please make sure you have enough space",
            memory_convert_bytes,
        )


def _check_memory_logits(sequences_list: List[str], vocab_size: int, pass_mode: str):
    """Function to compute the memory taken by the logits with float64 number.

    Args:
        sequences_list (str): sequences of proteins
        vocab_size (int]): Size of the vocabulary
        pass_mode (str): 'forward' or 'masked'
    """
    num_of_sequences = len(sequences_list)
    sum_seq_len = sum([len(seq) for seq in sequences_list])
    max_seq_len = max([len(seq) for seq in sequences_list])
    tensor_memory_bits = 64  # double/float64
    if pass_mode == "masked":
        memory_bits = sum_seq_len * max_seq_len * vocab_size * tensor_memory_bits
    elif pass_mode == "forward":
        memory_bits = num_of_sequences * max_seq_len * vocab_size * tensor_memory_bits

    memory_bytes = int(memory_bits / 8)
    memory_convert_bytes, is_warning = convert_bytes_size(memory_bytes)

    if is_warning:
        log.warning(
            "%s mode will need about %s of memory. Please make sure you have enough space",
            pass_mode,
            memory_convert_bytes,
        )


def _check_sequence(sequences_list: List[str], model: str, length: int):
    """Function that control sequence length

    Args:
        model : name of the model
        length : length limit to consider
    Raises:
        ValueError is model esm1b_t33_650M_UR50S and sequence_length >1024
    """
    if model == "esm1b_t33_650M_UR50S":
        is_longer = list(map(lambda x: len(x) > length, sequences_list))
        if sum(is_longer) > 0:
            raise ValueError(
                f"You cant't pass sequence with length more than {length} "
                f"with esm1b_t33_650M_UR50S, use esm1_t34_670M_UR100 or "
                f"filter the sequence length"
            )


def _check_tokens_list(sequences_list: List[str], tokens_list: List[str]):
    """Function that check if the list of tokens contains at least the tokens
    that are in the sequences.

    Args:
        sequences_list : list of sequences
        tokens_list : list of tokens to consider
    Raises:
        ValueError if some tokens in the sequences are not in the tokens_list
    """
    tokens = []
    for sequence in sequences_list:
        tokens += list(sequence)
        tokens = list(set(tokens))
    for token in tokens:
        if token not in tokens_list:
            raise ValueError(
                f"Token {token} is present in the sequences but not in the tokens_list."
            )


def _check_batch_size(batch_size: int, num_gpus: int):
    if not isinstance(batch_size, int):
        raise TypeError("batch_size should be of type int")
    if num_gpus > 1:
        if batch_size < num_gpus:
            raise ValueError(
                "With num_gpus>1, batch_size should be at least equal to num_gpus."
            )


def _get_num_batch_iter(model_inputs: Dict[str, Any], batch_size: int) -> int:
    """
    Get the number of batches when spliting model_inputs into chunks
    of size batch_size.
    """
    num_of_sequences = model_inputs["input_ids"].shape[0]
    num_batch_iter = int(np.ceil(num_of_sequences / batch_size))
    return num_batch_iter


def _generate_chunks(
    model_inputs: Dict[str, Any], batch_size: int
) -> Generator[Dict[str, Iterable], None, None]:
    """Yield a dictionnary of tensor"""
    num_of_sequences = model_inputs["input_ids"].shape[0]
    for i in range(0, num_of_sequences, batch_size):
        batch_sequence = {
            key: value[i : (i + batch_size)] for key, value in model_inputs.items()
        }
        yield batch_sequence


def load_fasta(path_fasta: Union[str, Path]) -> List[str]:
    """Read and parse records from a fasta file

    Args:
        path_fasta: path of the fasta file

    Returns:
        List: List of sequences
    """
    if not isinstance(path_fasta, Path):
        path_fasta = Path(path_fasta).resolve()
    return [str(record.seq) for record in SeqIO.parse(str(path_fasta), format="fasta")]


def get_logs_version(path_logs):
    """Get last version of logs folder to save model inside

    Args:
        path_logs (str): path of the logs/experiments folder
    """
    version_num = None
    try:
        # Folder version organize like version_x >> catch the 'x' integer
        version = str(max([int(fold.split("_")[1]) for fold in os.listdir(path_logs)]))
        version_num = "version_" + version
    except Exception as e:
        log.debug("Found exception %s" % e)
        version_num = None
    return version_num


def format_backend(backend_list: List[str]) -> List[str]:
    """format of list to display"""
    return ["  *" + " " * 3 + model for model in backend_list]


def list_backend() -> None:
    """Get all possible backend for the model"""
    print(
        "Use backend in this list :\n\n",
        "\n".join(format_backend(BACKEND_LIST)),
        sep="",
    )



import re
from typing import Dict, List, Tuple

import numpy as np
from biotransformers.utils.constant import NATURAL_AAS_LIST

ProbTuple = Tuple[float, float]
TokenProbsDict = Dict[int, Dict[str, float]]
SequenceProbsList = List[TokenProbsDict]


class InvalidPositionStringError(Exception):
    """Raised when a position string is passed with incorrect format"""


def validate_position_str(position_str: str):
    """Checks positions str format"""
    regex = re.compile(r"[A-Z]{1}[0-9]{1,4}[A-Z]{1}", re.I)
    if not regex.match(position_str):
        raise InvalidPositionStringError(
            f"'{position_str}' is not a valid position string"
        )


class Mutation:
    """register a mutation from a string

    Args:
        mutation (str): string mutation format "A8U" -> "NativeIdMutant"
    """

    def __init__(self, mutation_str: str) -> None:
        validate_position_str(mutation_str)
        self.mutation_str = mutation_str
        self.mutation = mutation_str[-1]
        self.native = mutation_str[0]
        self.position = int(mutation_str[1:-1])

    def __repr__(self) -> str:
        return f"Mutation >> Native: {self.native} New: {self.mutation} at position {self.position}"

    def is_valid_mutation(self, sequence: str):
        """Check if mutation is valid for the sequence of AA
        Args:
            sequence (str): protein sequence string
        """
        if len(sequence) < self.position:
            raise ValueError(
                f"Sequence smaller than position {self.position} for mutation {self.mutation_str}"
            )
        if self.native != sequence[self.position - 1]:
            raise InvalidPositionStringError(
                f"'{self.native}' is not a valid native position for mutation {self.mutation_str}"
            )
        if self.mutation not in NATURAL_AAS_LIST:
            raise ValueError(
                f"New amino acid {self.mutation} is not a valid mutation for {self.mutation_str}"
            )
        return True

def get_list_probs(
    mutation_list: List[Tuple[Mutation]],
    mutate_probs: SequenceProbsList,
    length_mutations: List[int],
) -> Tuple[List[List[float]], List[List[float]]]:
    """This function build a list of mutate and native probabilities to compute
    the mutate_score. For each position in the mutate list, we catch the native probability
    and the mutate probability of this position. We do this for each sequence and return two
    lists : native_probs and mutate probs.

    Args:
        mutation_list (List[Mutation]): list with integer which are mutations
        mutate_probs (List[Dict[Any]]): probabilities for mutate sequence
        length_mutations (List[int]):  length of indivual mutation for each sequence
    """
    flat_mutation = [mut for tup in mutation_list for mut in tup]
    native_probs_list, mutate_probs_list = [], []
    for prob, mut in zip(mutate_probs, flat_mutation):
        native_probs_list.append(prob[mut.position - 1][mut.native])
        mutate_probs_list.append(prob[mut.position - 1][mut.mutation])
    return split_list(native_probs_list, length_mutations), split_list(
        mutate_probs_list, length_mutations
    )


def mutation_score(native_probs: List[float], mutate_probs: List[float]) -> float:
    """
    Compute mutate score based on Masked marginal probability
    Sum(log(p(xi=xi_mutate|x-M))-log(p(xi=xi_native|x-M))) over M (M s a mutation set)

    Args:
        native_probs (List[ProbTuple]): [description]
        mutate_probs (List[ProbTuple]): [description]

    Returns:
        List[float]: [description]
    """
    return np.sum(
        [np.log(m_p) - np.log(n_p) for m_p, n_p in zip(mutate_probs, native_probs)]
    )


def split_list(list_to_split: List, lengths_list: List) -> List[List]:  # type: ignore
    """split a list in sublist

    Args:
        list_to_split (List): native list
        lengths_list (List): length of each sublist

    Returns:
        [type]: List of sublist
    """
    assert len(list_to_split) == sum(
        lengths_list
    ), "Sum of sublist length is not valid."
    splitted_list = []
    count = 0
    for length in lengths_list:
        splitted_list.append(list_to_split[count : (count + length)])
        count += length
    return splitted_list



"""This module provides a ProgressBar that works with ray and tqdm.
Each Ray workers update the progress bar remotly.
"""
from asyncio import Event
from typing import Tuple

import ray

# For typing purposes
from ray.actor import ActorHandle
from tqdm import tqdm


@ray.remote
class ProgressBarActor:
    counter: int
    delta: int
    event: Event

    def __init__(self) -> None:
        self.counter = 0
        self.delta = 0
        self.event = Event()

    def update(self, num_items_completed: int) -> None:
        """Updates the ProgressBar with the incremental
        number of items that were just completed.
        """
        self.counter += num_items_completed
        self.delta += num_items_completed
        self.event.set()

    async def wait_for_update(self) -> Tuple[int, int]:
        """Blocking call.

        Waits until somebody calls `update`, then returns a tuple of
        the number of updates since the last call to
        `wait_for_update`, and the total number of completed items.
        """
        await self.event.wait()
        self.event.clear()
        saved_delta = self.delta
        self.delta = 0
        return saved_delta, self.counter

    def get_counter(self) -> int:
        """
        Returns the total number of complete items.
        """
        return self.counter


# Back on the local node, once you launch your remote Ray tasks, call
# `print_until_done`, which will feed everything back into a `tqdm` counter.


class ProgressBar:
    progress_actor: ActorHandle
    total: int
    description: str
    pbar: tqdm

    def __init__(self, total: int, description: str = ""):
        # Ray actors don't seem to play nice with mypy, generating
        # a spurious warning for the following line,
        # which we need to suppress. The code is fine.
        self.progress_actor = ProgressBarActor.remote()  # type: ignore
        self.total = total
        self.description = description

    @property
    def actor(self) -> ActorHandle:
        """Returns a reference to the remote `ProgressBarActor`.

        When you complete tasks, call `update` on the actor.
        """
        return self.progress_actor

    def print_until_done(self) -> None:
        """Blocking call.

        Do this after starting a series of remote Ray tasks, to which you've
        passed the actor handle. Each of them calls `update` on the actor.
        When the progress meter reaches 100%, this method returns.
        """
        pbar = tqdm(desc=self.description, total=self.total)
        while True:
            delta, counter = ray.get(self.actor.wait_for_update.remote())
            pbar.update(delta)
            if counter >= self.total:
                pbar.close()
                return



"""This module build a general logger module"""
import logging
import os


def logger(module_name: str) -> logging.Logger:
    """Configure the logger with formatter and handlers.

    The log level depends on the environment variable `BIO_LOG_LEVEL`.

    - 0: NOTSET, will be set to DEBUG
    - 1: DEBUG
    - 2: INFO (default)
    - 3: WARNING
    - 4: ERROR
    - 5: CRITICAL
    https://docs.python.org/3/library/logging.html#levels

    Args:
        module_name (str): module name

    Returns:
        [Logger]: instantiate logger object
    """
    if module_name.endswith("py"):
        module_name = os.path.splitext(module_name)[0]

    logger_ = logging.getLogger(module_name)
    logger_.propagate = False
    log_level = os.environ.get("BIO_LOG_LEVEL", "2")
    log_level_int = max(int(log_level) * 10, 10)
    logger_.setLevel(log_level_int)

    handler = logging.StreamHandler()
    formatter = logging.Formatter("%(levelname)s: %(message)s")
    handler.setFormatter(formatter)
    handler.setLevel(log_level_int)
    logger_.addHandler(handler)

    return logger_



def lr_update(
    num_updates: int,
    warmup_updates: int,
    warmup_init_lr: float,
    lr_step: float,
    decay_factor: float,
) -> float:
    """InverseSquareRootSchedule.

    https://github.com/pytorch/fairseq/blob/master/fairseq/optim/lr_scheduler/inverse_square_root_schedule.py#L32

    Args:
        num_updates: number of batches already used.
        warmup_updates: number of batch steps for warm up.
        warmup_init_lr: initial learning rate.
        lr_step: step for increasing learning rate during warm up.
        decay_factor: factor for decreasing learning rate after warm up.

    Returns:
        learning rate multiplicate factor
    """
    if num_updates < warmup_updates:
        lr = warmup_init_lr + num_updates * lr_step
    else:
        lr = decay_factor * num_updates ** -0.5
    if warmup_init_lr > 0:
        return lr / warmup_init_lr

    return 0



import functools
import warnings


def deprecated_alias(**aliases):
    def deco(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)

        return wrapper

    return deco


def rename_kwargs(func_name, kwargs, aliases):  # noqa
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise TypeError(
                    "{} received both {} and {}".format(func_name, alias, new)
                )
            warnings.warn(
                "{} is deprecated; use {}".format(alias, new), DeprecationWarning, 3
            )

            if alias == "device":
                if "cuda" in kwargs[alias]:
                    kwargs.pop(alias)
                    kwargs[new] = 1
                elif "cpu" in kwargs[alias]:
                    kwargs.pop(alias)
                    kwargs[new] = 0
                else:
                    kwargs[new] = kwargs.pop(alias)

            elif alias == "multi_gpu":
                kwargs.pop(alias)
            else:
                kwargs[new] = kwargs.pop(alias)



"""Test module for testing logits function"""
import pytest

test_params = [
    (1, "forward"),
    (2, "masked"),
    (10, "forward"),
]

test_params_fasta = [(2, "forward")]


@pytest.mark.parametrize("batch_size, pass_mode", test_params)
def test_logits_type(init_model, batch_size, sequences, pass_mode):
    test_trans = init_model
    logits = test_trans.compute_logits(
        sequences,
        batch_size=batch_size,
        pass_mode=pass_mode,
    )
    assert len(logits) == len(sequences)
    for logit, sequence in zip(logits, sequences):
        assert logit.shape[0] == len(sequence)


@pytest.mark.parametrize("batch_size, pass_mode", test_params_fasta)
def test_logits_type_fasta(
    init_model, batch_size, fasta_path, lengths_sequence_fasta, pass_mode
):
    test_trans = init_model
    logits = test_trans.compute_logits(
        fasta_path,
        batch_size=batch_size,
        pass_mode=pass_mode,
    )
    for logit, length in zip(logits, lengths_sequence_fasta):
        assert logit.shape[0] == length



"""Setup file"""

import pathlib
from typing import List

from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

with open("biotransformers/version.py") as v:
    exec(v.readline())


def read_requirements() -> List:
    with open("requirements.txt", "r+") as file:
        requirements = [line.strip() for line in file.readlines()]

    return requirements


DESCRIPTION = "Wrapper on top of ESM/Protbert model in order to easily work with protein embedding"


def make_install():
    """main install function"""

    setup_fn = setup(
        name="bio-transformers",
        license="Apache-2.0",
        version=VERSION,  # noqa
        description=DESCRIPTION,
        author="Instadeep",
        long_description=README,
        long_description_content_type="text/markdown",
        author_email="a.delfosse@instadeep.com",
        packages=find_packages(exclude=["test"]),
        classifiers=[
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3.7",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Software Development",
        ],
        install_requires=read_requirements(),
        include_package_data=True,
        zip_safe=False,
        python_requires=">=3.7",
    )

    return setup_fn


if __name__ == "__main__":
    make_install()
